"
 COPYRIGHT (c) 1989-93 by Claus Gittinger
              All Rights Reserved

 This software is furnished under a license and may be used
 only in accordance with the terms of that license and with the
 inclusion of the above copyright notice.   This software may not
 be provided or otherwise made available to, or used by, any
 other person.  No title to or ownership of the software is
 hereby transferred.
"

StandardSystemView subclass:#DebugView
       instanceVariableNames:'haveControl exitAction canContinue
                              contextView codeView
                              receiverInspector contextInspector
                              contextArray selectedContext
                              catchBlock grabber traceView tracing
                              bigStep steppedContextAddress'
       classVariableNames:'theOneAndOnlyDebugger theOneAndOnlyNotifier'
       poolDictionaries:''
       category:'Interface-Debugger'
!

DebugView comment:'

COPYRIGHT (c) 1989-93 by Claus Gittinger
              All Rights Reserved

this class implements a graphical debugger interface.
To get control over the workstation while debugging, I get an exclusive
connection to the display and dispatch on this one. This will be changed
using processes soon.

$Header: /cvs/stx/stx/libtool/DebugView.st,v 1.2 1993-10-13 01:01:32 claus Exp $

written spring/summer 89 by claus
'!

!DebugView class methodsFor:'instance creation'!

new
    "return a DebugView - return the standard debugger if it already
     exists"

    |debugger|

    theOneAndOnlyDebugger isNil ifTrue:[
        debugger := super on:ModalDisplay.
        debugger label:'Debugger'.
        debugger icon:(Form fromFile:'Debugger.xbm' resolution:100).
        theOneAndOnlyDebugger := debugger
    ].
    ^ theOneAndOnlyDebugger
!

newDebugger
    "force creation of a new debugger"

    theOneAndOnlyDebugger := nil

    "DebugView newDebugger"
!

enterWithMessage:aString
    "the standard way of entering the debugger - sent from Objects
     error- and halt messages"

    |aDebugger|

    StepInterruptPending := nil.
    aDebugger := (self new) label:aString.
    aDebugger enter.
    ^ nil

    "nil halt"
!

enter
    "another way of entering the debugger"

    |aDebugger|

    StepInterruptPending := nil.
    aDebugger := (self new) label:'Debugger'.
    aDebugger enter.
    ^ nil

    "Debugger enter"
! !

!DebugView methodsFor:'initialization'!

initialize
    |v panel hpanel bpanel abortButton continueButton stepButton sendButton|

    super initialize.

    font := font on:device.

    exitAction := nil.

    bpanel := HorizontalPanelView
                        origin:(0.0 @ 0.0)
                        extent:(1.0 @ (font height * 2))
                            in:self.
    bpanel layout:#left.

    abortButton := Button
                        label:(resources at:'abort')
                        action:[abortButton turnOffWithoutRedraw. self doAbort]
                        in:bpanel.
    continueButton := Button
                        label:(resources at:'continue')
                        action:[continueButton turnOffWithoutRedraw. self doContinue]
                        in:bpanel.
    stepButton := Button
                        label:(resources at:'step')
                        action:[stepButton turnOff. self doStep]
                        in:bpanel.
    sendButton := Button
                        label:(resources at:'send')
                        action:[sendButton turnOff. self doSend]
                        in:bpanel.

    panel := VariableVerticalPanel
                        origin:(0.0 @ bpanel height)
                        corner:(1.0 @ 1.0)
                            in:self.

    v := ScrollableView for:SelectionInListView in:panel.
    v origin:(0.0 @ 0.0) corner:(1.0 @ 0.25).

    contextView := v scrolledView.
    contextView action:[:lineNr | self showSelection:lineNr].

    v := ScrollableView for:CodeView in:panel.
    v origin:(0.0 @ 0.25) corner:(1.0 @ 0.75).

    codeView := v scrolledView.

    hpanel := VariableHorizontalPanel in:panel.
    hpanel origin:(0.0 @ 0.75) corner:(1.0 @ 1.0).

    receiverInspector := InspectorView
                                origin:(0.0 @ 0.0) corner:(0.5 @ 1.0)
                                    in:hpanel.

    contextInspector := ContextInspectorView
                                origin:(0.5 @ 0.0) corner:(1.0 @ 1.0)
                                    in:hpanel
!

initializeMiddleButtonMenu
    |labels|

    labels := resources array:#('continue'
                                'abort'
                                '-'
                                'step'
                                'trace step'
                                'send'
"
                                'trace'
                                'untrace'
"
                                '-'
                                'resume'
                                'restart'
                                '-'
"
                                'breakpoints'
                                '-'
"
                                'exit smalltalk'
                              ).

    contextView
        middleButtonMenu:(PopUpMenu 
                                labels:labels
                             selectors:#(doContinue
                                         doAbort
                                         nil
                                         doStep
                                         doTraceStep
                                         doSend
"
                                         doTrace
                                         doNoTrace
"
                                         nil
                                         doResume
                                         doRestart
                                         nil
"
                                         doBreakpoints
                                         nil
"
                                         doExit)
                                  receiver:self
                                       for:contextView)

!

addToCurrentProject
    "ignored here"

    ^ self
!

createOnTop
    ^ false "true"
! !

!DebugView methodsFor:'error handling'!

catch:aSymbol with:someArgument for:anObject
    "this one is sent when an error occurs while in the debugger -
     we dont want another debugger to come up - do we ?"

    anObject class name print.
    ' ' print.
    aSymbol print. '(' print. someArgument print.
    ') within debugger cought' printNewline.
    (aSymbol == #halt:) ifFalse:[
        catchBlock value
    ].
    ^ nil
! !

!DebugView methodsFor:'interrupt handling'!

stepInterrupt
    |where here|

    bigStep ifTrue:[
        where := thisContext.  
        where := where sender.
        where := where sender.     
        here := where.
        where address == steppedContextAddress  ifFalse:[
            "
             check if we are in a context below steppedContext
             (i.e. if steppedContext can be reached from
              interrupted context. Not using context-ref but its
              address to avoid creation of many useless contexts..)
            "

            [where notNil] whileTrue:[
                where address == steppedContextAddress ifTrue:[
                    tracing ifTrue:[
                        here printString printNewline
                    ].
                    where := nil. here := nil.
                    "yes - continue"
                    "scedule another stepInterrupt
                     - must flush caches since optimized methods not always
                     look for pending interrupts"
                    ObjectMemory flushInlineCaches.
                    StepInterruptPending := true.
                    InterruptPending := true.
                    ^ nil
                ].
                where := where sender
            ]
        ].
        bigStep := false.
        tracing := false.
        self label:'after step'
    ] ifFalse:[
        steppedContextAddress := nil.
        self label:'after send'
    ].
    "release refs to context"
    where := nil. here := nil.
    self enter
! !

!DebugView methodsFor:'basic'!

enter
    "enter the debugger - get and display the context, then start an
     exclusive event loop on top of eveything else"

    |where con oldRecursionLimit selection|

    oldRecursionLimit := RecursionLimit.
    RecursionLimit := nil.

    ActiveGrab notNil ifTrue:[
        grabber := ActiveGrab.
        ActiveGrab device ungrabPointer.
        ActiveGrab device synchronizeOutput.
        ActiveGrab := nil
    ] ifFalse:[
        grabber := nil
    ].

    self realize.
    self raise.
    Display synchronizeOutput.

    where := thisContext.
    where notNil ifTrue:[
        "where is <self enter> context"
        where := where sender.
        (where receiver == DebugView) ifTrue:[
            where := where sender
        ]
        "where is now interrupted methods context"
    ].
    self setContext:where.

    "select context, where halt was ..."
    ( #(halt error raise) includes:where sender selector) ifTrue:[
        selection := 3
    ] ifFalse:[
        ( #(halt: error:) includes:where selector) ifTrue:[
            selection := 2
        ]
    ].

    where := nil.

    "if we came here by a big-step, show the method where we are"
    steppedContextAddress notNil ifTrue:[
        selection := 3
    ].

    selection notNil ifTrue:[
        self showSelection:selection.
        contextView selection:selection
    ].

    canContinue := true.
    self controlLoop.

    (exitAction == #abort) ifTrue:[
        selectedContext := nil.
        1 to:contextArray size do:[:index |
            (#(doIt printIt inspectIt)
                includes:(contextArray at:index) selector) ifTrue:[
                selectedContext := contextArray at:index
            ]
        ].
        exitAction := #resume
    ].

    contextArray := nil.

    (exitAction == #step) ifFalse:[
        self unrealize.
        device synchronizeOutput.
        (exitAction == #resume) ifTrue:[
            selectedContext notNil ifTrue:[
                con := selectedContext.
                selectedContext := nil.
                ErrorActive := false.
                InInterrupt := nil.
                RecursionLimit := oldRecursionLimit.
                con unwind.
                'cannot resume selected context' printNewline
            ]
        ] ifFalse:[
            (exitAction == #restart) ifTrue:[
                selectedContext notNil ifTrue:[
                    con := selectedContext.
                    selectedContext := nil.
                    ErrorActive := false.
                    InInterrupt := nil.
                    RecursionLimit := oldRecursionLimit.
                    con restart.
                    'cannot restart selected context' printNewline
                ]
            ] ifFalse:[
                (exitAction == #terminate) ifTrue:[
                    selectedContext := nil.
                    ErrorActive := false.
                    InInterrupt := nil.
                    RecursionLimit := oldRecursionLimit.
                    Processor currentProcess terminate.
                    'cannot terminate process' printNewline
                ]
            ]
        ]
    ].

    selectedContext := nil.

    grabber notNil ifTrue:[
        grabber device grabPointerIn:(grabber id).
        ActiveGrab := grabber
    ].

    RecursionLimit := oldRecursionLimit.
    (exitAction == #step) ifTrue:[
        "scedule another stepInterrupt
         - must flush caches since optimized methods not always
         look for pending interrupts"
        ObjectMemory flushInlineCaches.

        ObjectMemory stepInterruptHandler:self.
        StepInterruptPending := true.
        InterruptPending := true
    ]
!

controlLoop
    "this is a kludge:
        start a dispatchloop which exits when
        either continue, resume or step is pressed
    "

    Smalltalk at:#ErrorHandler put:self.
    haveControl := true.
    [haveControl] whileTrue:[
        self controlLoopCatchingErrors
    ].
    catchBlock := nil.
    Smalltalk at:#ErrorHandler put:nil.

    codeView contents:nil.
    codeView acceptAction:nil.
    contextView contents:nil.
    receiverInspector noChoice.
    contextInspector noChoice
!

controlLoopCatchingErrors
    "setup a catch-block"
    catchBlock := [^ nil].

    "this is a kludge:
        we do not have multiple processes
        therefore we start another dispatch loop, which exits when
        either continue, resume or step is pressed
        or (via the catchBlock) if an error occures
    "
    device dispatchWhile:[haveControl]
!

setContext:aContext
    |con text
     index "{ Class: SmallInteger }" |

    aContext isNil ifTrue:[
        text := Array with:'** no context **'.
        contextArray := nil
    ] ifFalse:[
        con := aContext.
        index := 0.
        [con notNil] whileTrue:[
            index := index + 1.
            con := con sender
        ].
        text := Array new:index.
        contextArray := Array new:index.
        con := aContext.
        index := 1.
        [con notNil] whileTrue:[
            contextArray at:index put:con.
            text at:index put:(con printString).
            index := index + 1.
            con := con sender
        ]
    ].
    contextView list:text.
    receiverInspector noChoice.
    contextInspector noChoice
! !

!DebugView methodsFor:'user interaction'!

showSelection:lineNr
    "user clicked on a header line - show selected code in textView"

    |con homeContext sel method code
     implementorClass lineNrInMethod|

    contextArray notNil ifTrue:[
        con := contextArray at:lineNr.
        lineNrInMethod := con lineNumber.
        con isBlockContext ifTrue:[
            homeContext := con methodHome
        ] ifFalse:[
            homeContext := con
        ].
        sel := homeContext selector.
        sel notNil ifTrue:[
            implementorClass := homeContext searchClass 
                                        whichClassImplements:sel.
            implementorClass isNil ifTrue:[
                codeView contents:'** no method - no source **'
            ] ifFalse:[
                method := implementorClass compiledMethodAt:sel.
                code := method source.
                code isNil ifTrue:[
                    codeView contents:'** no source **'
                ]
            ].
            code isNil ifTrue:[
                codeView acceptAction:nil.
                contextInspector noChoice
            ] ifFalse:[
                codeView contents:code.
                (lineNrInMethod ~~ 0) ifTrue:[
                    lineNrInMethod > codeView list size ifTrue:[
                        lineNrInMethod := codeView list size + 1
                    ].
                    codeView selectLine:lineNrInMethod.
                    codeView makeSelectionVisible
                ].
                codeView acceptAction:[:code | self codeAccept:code]
            ].
            contextInspector inspect:con.
            receiverInspector inspect:homeContext receiver
        ].
        selectedContext := homeContext
    ].
    "clear out locals to prevent keeping around unneeded contexts (due to the
     block held in codeView).
     (not really needed, since stuff gets collected away sooner or later ..."

    con := nil.
    homeContext := nil
!

codeAccept:someCode
    "user wants some code to be recompiled - must unwind stack since everything above
     and including selected method cannot be continued"

    "walk up context chain and find highest context which is either the selected context,
     or - if its a block-context - whose home is the selected context"

    |con top sel implementorClass method newMethod|

    codeView cursor:Cursor execute.

    con := selectedContext.
    top := con.
    [con notNil] whileTrue:[
        (con methodHome == selectedContext) ifTrue:[
            top := con
        ].
        con := con sender
    ].
    "now, remove everything up to and including top from context chain"

"
    self setContext:(top sender).
"

    sel := selectedContext selector.
    implementorClass := selectedContext searchClass whichClassImplements:sel.
    method := implementorClass compiledMethodAt:sel.
    newMethod := Compiler compile:someCode
                         forClass:implementorClass
                       inCategory:(method category)
                        notifying:codeView.

    "if it worked, remove everything up to and including top
     from context chain"

    (newMethod notNil and:[newMethod ~~ #Error]) ifTrue:[
        self setContext:(top sender).

        "continue/step is no longer possible"
        canContinue := false.
        self showSelection:1.
        exitAction := #resume
    ].
    codeView cursor:Cursor normal
!

destroy
    "closing the debugger implies an abort"

    contextView middleButtonMenu hide.
    receiverInspector noChoice.
    contextInspector noChoice.
    self doAbort
!

doExit
    "exit from menu: immediate exit from smalltalk"

    Smalltalk exit
!

doBreakpoints
    ^ self
!

doSend
    "send from menu"

    canContinue ifTrue:[
        steppedContextAddress := nil.
        haveControl := false.
        exitAction := #step
    ]
!

doStep
    "step from menu"

    canContinue ifTrue:[
        selectedContext notNil ifTrue:[
            bigStep := true.
            steppedContextAddress := selectedContext address
        ] ifFalse:[
            bigStep := true.
            steppedContextAddress := (contextArray at:2) address
        ].
        haveControl := false.
        exitAction := #step
    ]
!

doTraceStep
    "tracestep from menu"

    canContinue ifTrue:[
        tracing := true.
        self doStep
    ]
!

doAbort
    "abort from menu"

    steppedContextAddress := nil.
    haveControl := false.
    Processor currentProcess id == 0 ifTrue:[
        "dont allow termination of main-thread"
        exitAction := #abort
    ] ifFalse:[
        exitAction := #terminate 
    ]
!

doResume
    "resume from menu"

    steppedContextAddress := nil.
    haveControl := false.
    exitAction := #resume
!

doRestart
    "restart from menu"

    steppedContextAddress := nil.
    haveControl := false.
    exitAction := #restart
!

doTrace
    |v b|

    traceView isNil ifTrue:[
        v := StandardSystemView on:Display.
        v label:'Debugger-Trace'.
        v icon:icon.

        b := Button label:'untrace' in:v.
        b origin:(0 @ 0) extent:(1.0 @ (b height)).
        b action:[
            StepInterruptPending := false.
            tracing := false.
            v unrealize.
            traceView := nil
        ].
        traceView := ScrollableView for:TextCollector in:v.
        traceView origin:(0 @ (b height))
                  extent:[v width @ (v height - b height)]
    ].
    v realize.
    tracing := true.
    self doStep
!

doNoTrace
    traceView notNil ifTrue:[
        traceView topView destroy.
        traceView := nil.
        tracing := false
    ]
!

doContinue
    "continue from menu"

    canContinue ifTrue:[
        steppedContextAddress := nil.
        tracing := false.
        haveControl := false.
        exitAction := #continue
    ]
! !
